/*

 File: mathexpr.tpl, 2008/02/11

 ACETAF, Version 3.0.1

 Copyright (C) Markus Neher, markus.neher@math.uni-karlsruhe.de
               Ingo Eble,    ingoeble@web.de

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "functions.tpl"

#include <algorithm>
#include <sstream>

/*-------------------------------------+
 | Definitions of the template-classes |
 | in file "MathExpr.h"                |
 +-------------------------------------*/

enum Token { PLUS    = '+' , 
	     MINUS   = '-' , 
	     MUL     = '*' ,
	     DIV     = '/' , 
	     LP      = '(' , 
	     RP      = ')' ,
	     SEP     = ','   };

template<class T> class in_range : public std::unary_function<T,bool> //is arg1 < x < arg2 ??
{
  T arg1, arg2;
  
public:
  
  explicit in_range(const T& x, const T& y) : arg1(x), arg2(y) {}
  bool operator()(const T& x) const { return std::less<T>()(arg1,x) && std::less<T>()(x,arg2); }
};

//Type expression

template<class T> MathExpr<T>::MathExpr(const std::string& ss,const std::string& var = "x") throw(syntax_error) : s(ss)
{
  if( !s.empty() )
    {
      /*
        Syntax check for ++,+-,-+,--
      */

      int check = 3 + s.find("++") + s.find("+-") + s.find("-+") + s.find("--");
      if( check >= 0 ) throw syntax_error("What does '" + s + "' mean?");

      int size( s.length() );
      std::vector<int> pos_plus_and_minus;
      in_range<int> is_in_range( -1 , size );

      int pos_plus( s.find( PLUS ) ), pos_minus( s.find( MINUS ) );

      while( is_in_range( pos_plus ) )// -1 < pos_plus < size or 0 <= pos_plus <= size-1 
	{
	  if( s[ pos_plus-1 ] != 'E' && s[ pos_plus-1 ] != 'e' )
	    {
	       pos_plus_and_minus.push_back( pos_plus );
	    }
          pos_plus = s.find( PLUS, pos_plus+1 );
	}

      while( is_in_range( pos_minus ) ) 
	{
	  if( s[ pos_minus-1 ] != 'E' && s[ pos_minus-1 ] != 'e' )
	    {
	       pos_plus_and_minus.push_back( pos_minus );
	    }
	  pos_minus = s.find( MINUS, pos_minus+1 );
	}

      if( pos_plus_and_minus.empty() )
	{ //e.g. "2*12", "exp(23/4)", etc.
	  terms.push_back( term<T>( s, var ) );
	}
      else
	{
	  typename std::vector<int>::iterator first = pos_plus_and_minus.begin(),
	                                      last  = pos_plus_and_minus.end();

	  std::sort( first, last, std::less<int>());

	  int pos_lp( s.find( LP ) ), pos_rp( s.find( RP ) );
	
	  while( is_in_range( pos_lp ) || is_in_range( pos_rp ) )
	    {
	      if( is_in_range( pos_lp ) ) 
		{
		  if( is_in_range( pos_rp ) )
		    {
		      if( pos_lp > pos_rp ) 
			{
			  throw syntax_error("Some brackets are in wrong order: ')..('.");
			}
		      else
			{
			  last = std::remove_if( first, last, in_range<int>(pos_lp, pos_rp) );
			  pos_lp = s.find( LP, pos_lp+1 );
			  pos_rp = s.find( RP, pos_rp+1 );
			}
		    }
		  else
		    {
		      throw syntax_error(" ')' missing.");
		    }
		}
	      else
		{
		  if( is_in_range( pos_rp ) ) 
		    {
		      throw syntax_error(" '(' missing.");
		    }
		}
	    }

	  if( first != last-- )
	    {	     		  
              if( *first != 0 ) //expression without a leading sign
                { //e.g. "1+.." or "5-.."
	          terms.push_back( term<T>( s.substr( 0, *first ), var ) );	      
                }
	      while( first != last )
		{
		  int pos( *first );
		  int length( *(++first) - pos );
		  terms.push_back( term<T>( s.substr( pos, length ), var ) );	      
		}
	      terms.push_back( term<T>( s.substr( *first, size-*first ), var ) );	      
	    }
	  else
	    { //e.g. "(2+1-34)", "sin(2-3*12)", etc.
	      terms.push_back( term<T>( s, var ) );
	    }
	}
    }
  else
    {
      std::cerr<<"No string available."<<std::endl;
    }
}


//Type term

template<class T> term<T>::term(const std::string& s, const std::string& var) throw(syntax_error)
{
  if( !s.empty() )
    {
      int size( s.length() );
      std::vector<int> pos_mul_and_div;
      in_range<int> is_in_range( -1 , size );

      int pos_mul( s.find( MUL ) ), pos_div( s.find( DIV ) );

      while( is_in_range( pos_mul ) )// -1 < pos_mul < size or 0 <= pos_mul <= size-1 
	{
	  pos_mul_and_div.push_back( pos_mul );
	  pos_mul = s.find( MUL, pos_mul+1 );
	}

      while( is_in_range( pos_div ) ) 
	{
	  pos_mul_and_div.push_back( pos_div );
	  pos_div = s.find( DIV, pos_div+1 );
	}

      if( pos_mul_and_div.empty() )
	{ // e.g. "(1+2-2.2334)", "exp(1+2)", etc.
	  prims.push_back( prim<T>( s, var ) );
	}
      else
	{
	  typename std::vector<int>::iterator first = pos_mul_and_div.begin(),
	                                      last  = pos_mul_and_div.end();

	  std::sort( first, last, std::less<int>());

	  int pos_lp( s.find( LP ) ), pos_rp( s.find( RP ) );
	
	  while( is_in_range( pos_lp ) || is_in_range( pos_rp ) )
	    {
	      if( is_in_range( pos_lp ) ) 
		{
		  if( is_in_range( pos_rp ) )
		    {
		      if( pos_lp > pos_rp ) 
			{ //e.g. "2*)3/2(", ")1(", etc.
			  throw syntax_error("Some brackets are in wrong order: ')..('.");
			}
		      else
			{
			  last = std::remove_if( first, last, in_range<int>(pos_lp, pos_rp) );
			  pos_lp = s.find( LP, pos_lp+1 );
			  pos_rp = s.find( RP, pos_rp+1 );
			}
		    }
		  else
		    { //e.g. "(1*2", "12/(2", etc.
		      throw syntax_error(" ')' missing.");
		    }
		}
	      else
		{
		  if( is_in_range( pos_rp ) ) 
		    { //e.g. "1*2)", "12/2)", etc.
		      throw syntax_error(" '(' missing.");
		    }
		}
	    }

	  if( first != last-- )
	    {	     		  
              if( *first == 0 ) //term with a leading operator * or /
                { //e.g. "*12.3" or "/sin(2.1)"
                  throw syntax_error("First operand missing in '" + s + "'.");
                }
	      prims.push_back( prim<T>( s.substr( 0, *first ), var ) );	      
	      while( first != last )
		{
		  int pos( *first );
		  int length( *(++first) - pos );
		  prims.push_back( prim<T>( s.substr( pos, length ), var ) );	      
		}
	      prims.push_back( prim<T>( s.substr( *first, size-*first ), var ) );	      
	    }
	  else
	    { //e.g. "(2*12+2*x)", "sin(1+2/3)", etc.
	      prims.push_back( prim<T>( s, var ) );
	    }
	}
    }
  else
    {
      std::cerr<<"No string available."<<std::endl;
    }
}

//Type prim

//The following function is needed for evaluation of a complex expression
//and will be defined later for particular types.

template<class T> inline T* set_imag()
{
}

template<class T> prim<T>::prim(const std::string& s, const std::string& var) throw(syntax_error)
{
  if( !s.empty() )
    {
      std::string name("");
      std::string::const_iterator first = s.begin(),
                                  last  = s.end();
      OP = VZ = false;//default: OP = '*' and VZ = '+'
      fp1 = 0; fp2 = 0; fp3 = 0; expr = 0; return_value = 0;

      while( first != last )
	{
	  char ch( *first++ );

	  switch( ch )
	    {
	    case PLUS: break;
	    case MINUS:
	      {
		VZ = !VZ;//because: (-1)*(-1) = +1 and (-1)*(-1)*(-1) = -1 and so on ...
		break;
	      }
	    case MUL: break;
	    case DIV:
	      {
		OP = true;
		break;
	      }
	    case LP:
	      { // "Token()" inserted 080211
		if( Token(*(--last)) != RP ) //the case: e.g. "(2*3+4)!" or "(2+1)" or "(4/(2+3))sin"
		  throw syntax_error("What does '" + s + "' mean?");

		std::string content( first, last );
	        if( fp1 || fp2 || fp3 )
		  { //content is an argument list of a function
		    if( content.empty() ) //the case: e.g. "sin()" or "exp()"
		      throw syntax_error("No argument(s) in function '" + name + " '.");

		    int pos_sep( content.find( SEP ) );
		    bool is_sep_in_range( in_range<int>( -1 , content.length() )( pos_sep ) );
		    if( fp1 )
		      { //function with one argument
			if( is_sep_in_range ) //the case: e.g. "sin(12+3,4)" or "sin(,12)"
			  throw syntax_error("In function '" + name + " ', only one argument is allowed.");

			else
			  {
			    fp1->set_arg( MathExpr<T>( content, var ) );
			    first = last;//leave the loop
			  }
		      }
		    else
		      { //function with two arguments
			if( !is_sep_in_range ) //the case: e.g. "root(12)" or "power(2*x)"
			  throw syntax_error("In function '" + name + " ', two arguments are required.");

			else 
			  {
			    if( pos_sep == 0 ) //the case: e.g. "root(,2)"
			      throw syntax_error("In function '" + name + " ', the first argument is missing.");
			    if( pos_sep == content.length()-1 ) //the case: e.g. "root(12*x,)"
			      throw syntax_error("In function '" + name + " ', the second argument is missing.");
			    if( fp2 ) fp2->set_1st_arg( MathExpr<T>( content.substr( 0, pos_sep ), var ) );
			    if( fp3 ) fp3->set_1st_arg( MathExpr<T>( content.substr( 0, pos_sep ), var ) );
			    first += pos_sep;
			  }
		      }
		  }
		else
		  { //content is a simple expression in brackets
		    if( content.empty() ) //the case: "()"
		      throw syntax_error("Empty brackets found.");

		    expr = new MathExpr<T>( content, var );
		    first = last;//leave the loop
		  }
		break;
	      }
	    case SEP:
	      {
		if( fp2 || fp3 ) 
		  {
		    if( fp2 ) fp2->set_2nd_arg( MathExpr<int>( std::string( first, last ), var ) );
		    if( fp3 ) fp3->set_2nd_arg( MathExpr<T>  ( std::string( first, last ), var ) );
		    first = last;//leave the loop
		  }
		else //the case: e.g. ",sin(12)" or ",2.123"
		  throw syntax_error("What is the meaning of '" + s + "'?");

		break;
	      }
	    case '0': case '1': case '2': case '3': case '4':
	    case '5': case '6': case '7': case '8': case '9':
	    case '.': 
	      {
		std::string number( --first, last );

		if( isalpha( number.at( number.size()-1 ) ) )
		  throw syntax_error("What is the meaning of '" + number + "'?");

		return_value = new T;

		number = (number + "," + number);

		number >> (*return_value);

		if( OP && is_zero<T>()( *return_value ) ) throw syntax_error("Division by zero.");
		first = last;//leave the loop
		break;
	      }
	    default:
	      {
		if( isalpha(ch) ) 
		  { 
		    name += ch;
		    while ( first != last && (ch = *first++) && isalnum(ch) ) 
		      {
			name += ch;
		      }

		    if( Token(ch) == LP )  // "Token()" inserted 080211
		      { // Set function_pointer. Puuhh!
			first--;
			if      ( name == "sin"   ) fp1 = new function_T<T> ( SIN<T>   );
			else if ( name == "cos"   ) fp1 = new function_T<T> ( COS<T>   );
			else if ( name == "tan"   ) fp1 = new function_T<T> ( TAN<T>   );
			else if ( name == "cot"   ) fp1 = new function_T<T> ( COT<T>   );
			else if ( name == "acos"  ) fp1 = new function_T<T> ( ACOS<T>  );
			else if ( name == "asin"  ) fp1 = new function_T<T> ( ASIN<T>  );
			else if ( name == "atan"  ) fp1 = new function_T<T> ( ATAN<T>  );
			else if ( name == "acot"  ) fp1 = new function_T<T> ( ACOT<T>  );
			else if ( name == "cosh"  ) fp1 = new function_T<T> ( COSH<T>  );
			else if ( name == "sinh"  ) fp1 = new function_T<T> ( SINH<T>  );
			else if ( name == "tanh"  ) fp1 = new function_T<T> ( TANH<T>  );
			else if ( name == "coth"  ) fp1 = new function_T<T> ( COTH<T>  );
			else if ( name == "acosh" ) fp1 = new function_T<T> ( ACOSH<T> );
			else if ( name == "asinh" ) fp1 = new function_T<T> ( ASINH<T> );
			else if ( name == "atanh" ) fp1 = new function_T<T> ( ATANH<T> );
			else if ( name == "acoth" ) fp1 = new function_T<T> ( ACOTH<T> );
			else if ( name == "exp"   ) fp1 = new function_T<T> ( EXP<T>   );
			else if ( name == "ln"    ) fp1 = new function_T<T> ( LN<T>    );
			else if ( name == "sqr"   ) fp1 = new function_T<T> ( SQR<T>   );
			else if ( name == "sqrt"  ) fp1 = new function_T<T> ( SQRT<T>  );
			else if ( name == "power" ) fp2 = new function_Ti<T>( POWER<T> );
			else if ( name == "root"  ) fp2 = new function_Ti<T>( SQRT<T>  );
			else if ( name == "pow"   ) fp3 = new function_TT<T>( POW<T>   );
			else throw syntax_error("Function '" + name + "' unknown.");
		      }
		    else 
		      { //Atom is not a function! Check if it is the free variable or the imaginary unit.
			if( name == "i" )
			  {
			    return_value = set_imag<T>();
			  }
			else if( name != var ) throw syntax_error("'" + name + "' unknown.");
			first = last;//quit the loop
		      }
		  }
		else 
		  {
		    throw syntax_error("Expression not finished.");
		    break;
		  }
	      }
	    }
	}
    }
  else
    {
      std::cerr<<"No string available."<<std::endl;
    }
}					 

//Specialization of template prim<T> for type T = int
template<> prim<int>::prim(const std::string&, const std::string&) throw(syntax_error);

/*

  End of File: mathexpr.tpl

*/
